#include <iostream>
#include <vector>
#include "isaac/isaac64.h"
#include "Poly.cpp"
#include "Setup.cpp"

using namespace std;

extern const int d;

template<class T>
class Matrix {
    friend Matrix operator+(const Matrix& lhs, const Matrix& rhs) {
        if (lhs._cols != rhs._cols || lhs._rows != rhs._rows) {
            throw "Size mismatch";
        }
        Matrix<T> res(lhs._rows, lhs._cols);
        for (int i = 0; i < lhs._rows; i++) {
            for (int j = 0; j < lhs._cols; j++) {
                res(i, j) = lhs(i, j) + rhs(i, j);
            }
        }
        return res;
    }
    friend Matrix operator-(const Matrix& lhs, const Matrix& rhs) {
        if (lhs._cols != rhs._cols || lhs._rows != rhs._rows) {
            throw "Size mismatch";
        }
        Matrix<T> res(lhs._rows, lhs._cols);
        for (int i = 0; i < lhs._rows; i++) {
            for (int j = 0; j < lhs._cols; j++) {
                res(i, j) = lhs(i, j) - rhs(i, j);
            }
        }
        return res;
    }

    template<class S>
    friend Matrix operator*(S scalar, const Matrix& rhs) {
        Matrix<T> res(rhs._rows, rhs._cols);
        for (int i = 0; i < rhs._rows; i++) {
            for (int j = 0; j < rhs._cols; j++) {
                res(i, j) = scalar * rhs(i, j);
            }
        }
        return res;
    }

    friend Matrix operator*(const Matrix& lhs, const Matrix& rhs) {
        //cout << "(" << lhs._rows << "," << lhs._cols << ")x(" << rhs._rows << "," << rhs._cols << ")" << endl;
        if (lhs._cols != rhs._rows) throw "Size mismatch";
        Matrix<T> res(lhs._rows, rhs._cols);
        for (int i = 0; i < res._rows; i++) {
            for (int j = 0; j < res._cols; j++) {
                vector<T> sum(8);
#pragma omp parallel for
                for (int k = 0; k < lhs._cols; k++) {
                    sum[omp_get_thread_num()] += lhs(i, k) * rhs(k, j);
                }
                T totalsum = T(0);
                for (int k = 0; k < sum.size(); k++) totalsum += sum[k];
                res(i, j) = totalsum;
            }
        }
        return res;
        Matrix<array<Z, primrootpow2>> a(lhs.rows(), lhs.cols()), b(rhs.rows(), rhs.cols());
#pragma omp parallel for collapse(2)
        for (int i = 0; i < lhs.rows(); i++) {
            for (int j = 0; j < lhs.cols(); j++) {
                for (int k = 0; k < d; k++) {
                    a(i, j)[k] = Montgomery::augment(lhs(i, j)[k]);
                }
                Poly<d, Z>::DFT(a(i, j), false);
            }
        }
#pragma omp parallel for collapse(2)
        for (int i = 0; i < rhs.rows(); i++) {
            for (int j = 0; j < rhs.cols(); j++) {
                for (int k = 0; k < d; k++) {
                    b(i, j)[k] = Montgomery::augment(rhs(i, j)[k]);
                }
                Poly<d, Z>::DFT(b(i, j), false);
            }
        }
        for (int i = 0; i < res._rows; i++) {
            for (int j = 0; j < res._cols; j++) {
                vector<T> sum(8);
#pragma omp parallel for
                for (int k = 0; k < lhs._cols; k++) {
                    sum[omp_get_thread_num()] += Poly<d, Z>::multFFTaugmented(a(i, k), b(k, j));
                }
                T totalsum = T(0);
                for (int k = 0; k < sum.size(); k++) totalsum += sum[k];
                res(i, j) = totalsum;
            }
        }
        return res;
    }

    template<class S>
    friend Matrix operator%(const Matrix& lhs, const S modulo) {
        Matrix<T> res(lhs._rows, lhs._cols);
#pragma omp parallel for
        for (int i = 0; i < res.size(); i++) {
            res[i] = lhs[i] % modulo;
        }
        return res;
    }

    friend Matrix operator|(const Matrix& lhs, const Matrix& rhs) {
        if (lhs._rows != rhs._rows) throw "Size mismath";
        Matrix<T> res(lhs._rows, lhs._cols + rhs._cols);
        for (int i = 0; i < lhs._rows; i++) {
            for (int j = 0; j < lhs._cols; j++) res(i, j) = lhs(i, j);
            for (int j = 0; j < rhs._cols; j++) res(i, j + lhs._cols) = rhs(i, j);
        }
        return res;
    }
    friend ostream& operator<<(ostream& os, const Matrix& rhs)
    {
        for (int i = 0; i < rhs._rows; i++) {
            for (int j = 0; j < rhs._cols; j++) {
                os << rhs(i, j) << "   ";
            }
            os << endl;
        }
        return os;
    }
protected:
    std::vector<T> elems;
    size_t _rows, _cols;
public:
    Matrix() : elems(0) {
        _rows = 0;
        _cols = 0;
    }
    Matrix(size_t rows, size_t cols) : elems(rows* cols) {
        _rows = rows;
        _cols = cols;
    }
    Matrix operator-() {
        Matrix<T> res(this->_rows, this->_cols);
        for (size_t i = 0; i < res.size(); i++) {
            res[i] = -(*this)[i];
        }
        return res;
    }
    T& operator[](size_t e) {
        return elems[e];
    }
    T operator[](size_t e) const {
        return elems[e];
    }
    T& operator()(size_t row, size_t col) {
        return elems[row * _cols + col];
    }
    T operator()(size_t row, size_t col) const {
        return elems[row * _cols + col];
    }
    size_t size() const {
        return _rows * _cols;
    }
    size_t rows() const {
        return _rows;
    }
    size_t cols() const {
        return _cols;
    }
    Matrix range(int row_start, int col_start, int row_end, int col_end) {
        if (row_start < 0) row_start = this->_rows + row_start;
        if (row_end < 0) row_end = this->_rows + row_end;
        if (col_start < 0) col_start = this->_cols + col_start;
        if (col_end < 0) col_end = this->_cols + col_end;

        Matrix<T> res(row_end - row_start + 1, col_end - col_start + 1);
        for (size_t i = row_start; i <= row_end; i++) {
            for (size_t j = col_start; j <= col_end; j++) {
                res(i - row_start, j - col_start) = (*this)(i, j);
            }
        }
        return res;
    }
    Matrix transpose() {
        Matrix<T> res(this->_cols, this->_rows);
        for (int i = 0; i < this->_rows; i++) {
            for (int j = 0; j < this->_cols; j++) {
                res(j, i) = (*this)(i, j);
            }
        }
        return res;
    }
    T dot(const Matrix& rhs) {
        if (this->size() != rhs.size()) throw "Size mismath";
        T sum = T(0);
        for (int i = 0; i < this->size(); i++) {
            sum += (*this)[i] * rhs[i];
        }
        return sum;
    }
    Matrix outer(const Matrix& rhs) {
        Matrix<T> res(this->size(), rhs.size());

#pragma omp parallel for
        for (int i = 0; i < res.rows(); i++) {
            for (int j = 0; j < res.cols(); j++) {
                res(i, j) = (*this)[i] * rhs[j];
            }
        }
        return res;
    }
    Matrix addToFirstRow(const Matrix& rhs) {
        if (_cols != rhs.size()) {
            throw "Size mismatch";
        }
        Matrix<T> res(_rows, _cols);
        for (int j = 0; j < _cols; j++) {
            res(0, j) = (*this)(0, j) + rhs[j];
        }
        for (int i = 1; i < _rows; i++) {
            for (int j = 0; j < _cols; j++) {
                res(i, j) = (*this)(i, j);
            }
        }
        return res;
    }
    Matrix randomRowSum() {
        Matrix<T> sum(1, this->cols());
        for (int i = 0; i < this->rows(); i++) {
            if ((irand64() & 1) == 1) {
                for (int j = 0; j < this->cols(); j++) {
                    sum(0, j) += (*this)(i, j);
                }
            }
        }
        return sum;
    }
    Matrix Maugment() {
        Matrix<T> res(this->_rows, this->_cols);
        for (int i = 0; i < this->size(); i++) {
            res[i] = elems[i].Maugment();
        }
        return res;
    }
    Matrix Mreduce() {
        Matrix<T> res(this->_rows, this->_cols);
        for (int i = 0; i < this->size(); i++) {
            res[i] = elems[i].Mreduce();
        }
        return res;
    }
    template <typename F>
    static Matrix Random(F r, size_t rows, size_t cols) {
        Matrix<T> res(rows, cols);
        for (int i = 0; i < res.size(); i++) {
            res[i] = r();
        }
        return res;
    }
    template<typename S>
    static Matrix Uniform(S q, size_t rows, size_t cols) {
        return Random([q]() {return T::Uniform(); }, rows, cols);
    }
    static Matrix Chi(size_t rows, size_t cols) {
        return Random(T::Chi, rows, cols);
    }
};
