#ifndef MONTGOMERY_H 
#define MONTGOMERY_H 

#include <iostream>
#include <type_traits>
#include <boost/multiprecision/cpp_int.hpp>
#include "Setup.cpp"
#include "PrintInt128.cpp"
#ifdef _MSC_VER
#include <intrin.h>
#endif
using namespace std;
typedef Z T;
typedef uZ uT;

class Montgomery {
protected:
	static uT _n;
	static int _d;
	static uT _np;
	static uT _r2;
public:
	inline static T mul_high_half(uint64_t x, uint64_t y) {
		uint32_t a = x >> 32, b = x;
		uint32_t c = y >> 32, d = y;
		uint64_t ac = (uint64_t)a * c;
		uint64_t ad = (uint64_t)a * d;
		uint64_t bc = (uint64_t)b * c;
		uint64_t bd = (uint64_t)b * d;
		uint64_t carry = (uint64_t)(uint32_t)ad + (uint64_t)(uint32_t)bc + (bd >> 32u);
		uint64_t high = ac + (ad >> 32u) + (bc >> 32u) + (carry >> 32u);
		return high;
	}
#ifdef __GNUC__
	typedef unsigned __int128 uint128;
	typedef __int128 int128;
	inline static void mult_full(uint64_t x, uint64_t y, uint64_t& high, uint64_t& low) {
		uint128 prod = (int128)x * y;
		high = (uint64_t)(prod >> 64);
		low = (uint64_t)prod;
	}
	inline static void mult_full(int128 x, int128 y, int128& high, int128& low) {
		counter++;
		uint64_t a = x >> 64, b = x;
		uint64_t c = y >> 64, d = y;
		uint128 ac = (uint128)a * c;
		uint128 ad = (uint128)a * d;
		uint128 bc = (uint128)b * c;
		uint128 bd = (uint128)b * d;
		uint128 carry = (uint128)(uint64_t)ad + (uint128)(uint64_t)bc + (bd >> 64u);
		high = ac + (ad >> 64u) + (bc >> 64u) + (carry >> 64u);
		low = (ad << 64u) + (bc << 64u) + bd;
	}
#else
	inline static void mult_full(uint64_t x, uint64_t y, uint64_t& high, uint64_t& low) {
		high = __umulh(x, y);
		low = x * y;
	}
#endif
	typedef boost::multiprecision::int128_t b128;
	typedef boost::multiprecision::int256_t b256;
	typedef boost::multiprecision::int512_t b512;
	inline static void mult_full(b128 x, b128 y, b128& high, b128& low) {
		b256 res = boost::multiprecision::multiply(res, x, y);
		low = static_cast<b128>(res);
		res >>= 128;
		high = static_cast<b128>(res);
	}
	inline static void mult_full(b256 x, b256 y, b256& high, b256& low) {
		b512 res = boost::multiprecision::multiply(res, x, y);
		low = static_cast<b256>(res);
		res >>= 256;
		high = static_cast<b256>(res);
	}


	static void set_mod(T n) {
		if (n == _n) return;
		_n = static_cast<uT>(n);
		_d = sizeof(T) * 8;
		if (n == 0) {
			_np = 0;
			_r2 = 0;
			return;
		}
		uT inv = 1;
		for (int i = 0; i < 7; i++)
			inv *= 2 - _n * inv;
		_np = inv;
		//std::cout << inv << std::endl;

		_r2 = (1 - _n) % _n - 1;
		//cout << _r2 << endl;
		for (int i = 0; i < 4; i++) {
			_r2 <<= 1;
			if (_r2 >= _n)
				_r2 -= _n;

		}
		for (int i = 0; i < log2(sizeZ) + 1; i++)
			_r2 = static_cast<uT>(mul(_r2, _r2));
		//std::cout << _r2 << std::endl;
	}
	inline static T augment(T x) {
		//x %= _n;
		//cout << x << endl;
		if (x < 0) x += _n;
		return mul((T)x, _r2);
	}
	inline static T reduce(T x) {
		if (x < 0) x += _n;
		T high, low;
		T m = x * _np;
		mult_full(m, _n, high, low);
		T t = high;
		if (t == 0) return 0;
		else return _n - t;
	}
	inline static T reduce(T high, T low) {
		T high2, low2;
		T m = low * _np;
		mult_full(m, _n, high2, low2);
		T t = high - high2;

		if (t < 0) return t + _n;
		else return t;
	}
	inline static T mul(T x, T y) {
		T high, low, high2, low2;

		mult_full(x, y, high, low);

		T m = low * _np;
		mult_full(m, _n, high2, low2);
		T t = high - high2;

		if (t < 0) return t + _n;
		else return t;
	}
	inline static T add(T x, T y) {
		T s = x + y;
		if (s >= _n) s -= _n;
		return s;
	}
	inline static T add(T x, T y, T z) {
		T s = x + y + x;
		if (s >= _n) s -= _n;
		return s;
	}
	inline static T sub(T x, T y) {
		if (y > x) return _n + x - y;
		else return x - y;
	}
	static T modPow(T b, T e) {
		T res = augment((T)1);
		T bM = augment(b);
		while (e > 0) {
			if (e % 2 == 1) {
				res = mul(res, bM);
			}
			e >>= 1;
			bM = mul(bM, bM);
		}
		return reduce(res);
	}
	static T inv(T x) {
		return modPow(x, _n - 2);
	}
	static uint64_t counter;
};

uT Montgomery::_n;
uT Montgomery::_np;

int Montgomery::_d;
uT Montgomery::_r2 = 0;

uint64_t Montgomery::counter = 0;

#endif