#ifndef POLY_H
#define POLY_H

#include <iostream>
#include <array>
#include <vector>
#include <stack>
#include <random>
#include <initializer_list>
//#include <type_traits>
#include <boost/multiprecision/cpp_int.hpp>
#include <boost/type_traits.hpp>

#include "Montgomery.cpp"
#include "isaac/isaac64.h"
#include "RandomDist.cpp"
#include "Setup.cpp"

using namespace std;

Z primroot;
Z primrootinv;

vector<int> phi625 = { 0, 125, 250, 375, 500 };

template<size_t d, class T>
class Poly
{
    typedef uZ uT;
    friend Poly operator+(const Poly& lhs, const Poly& rhs)
    {
        Poly<d, T> res;
        if (_q == 0)
            for (int i = 0; i < d; i++)
                res[i] = lhs[i] + rhs[i];
        else
            for (int i = 0; i < d; i++) {
                res[i] = lhs[i] + rhs[i];
                if (res[i] > _q) res[i] -= _q;
                else if (res[i] < -_q) res[i] += _q;
                res[i] = Montgomery::add(lhs[i], rhs[i]);
            }
        return res;
    }
    friend Poly operator-(const Poly& lhs, const Poly& rhs)
    {
        Poly<d, T> res;
        if (_q == 0)
            for (int i = 0; i < d; i++)
                res[i] = lhs[i] - rhs[i];
        else
            for (int i = 0; i < d; i++)
                res[i] = Montgomery::sub(lhs[i], rhs[i]);
        return res;
    }
    friend Poly operator*(T scalar, const Poly& rhs) {
        Poly<d, T> res;
        if (_q == 0) {
            for (int i = 0; i < d; i++) {
                res[i] = scalar * rhs[i];
            }
        }
        else {
            Poly<d, T> rhsM;
            for (int i = 0; i < d; i++) {rhsM[i] = Montgomery::augment(rhs[i]);}
            T Mscalar = Montgomery::augment(scalar);
            for (int i = 0; i < d; i++) {
                res[i] = Montgomery::mul(Mscalar, rhsM[i]);
            }
            for (int i = 0; i < d; i++) { res[i] = Montgomery::reduce(res[i]); }
        }
        return res;
    }
    friend Poly operator/(const Poly& lhs, T scalar) {
        Poly<d, T> res;
        for (int i = 0; i < d; i++) {
            res[i] = lhs[i] / scalar;
        }
        return res;
    }
    friend Poly operator*(const Poly& lhs, const Poly& rhs) {
        Poly<d, T> res; 
        if (_q == 0) {
            for (int i = 0; i < d; i++) {
                for (int j = 0; j < d; j++) {
                    if (j < i) {
                        res[j] -= lhs[i] * rhs[d - i + j];
                    }
                    else {
                        res[j] += lhs[i] * rhs[j - i];
                    }
                }
            }
        }
        else {
            bool useFFT = true;
            if (useFFT) {
                return multFFT(lhs, rhs);
            }
            else {
                Poly<d, T> lhsM, rhsM;
                for (int i = 0; i < d; i++) {
                    lhsM[i] = Montgomery::augment(lhs[i]);
                    rhsM[i] = Montgomery::augment(rhs[i]);
                }
                for (int i = 0; i < d; i++) {
                    for (int j = 0; j < d; j++) {
                        if (j < i) {
                            res[j] = Montgomery::sub(res[j], Montgomery::mul(lhsM[i], rhsM[d - i + j]));
                        }
                        else {
                            res[j] = Montgomery::add(res[j], Montgomery::mul(lhsM[i], rhsM[j - i]));
                        }
                    }
                }
                //cout << res << endl;
                for (int i = 0; i < d; i++) {
                    res[i] = Montgomery::reduce(res[i]);
                }
            }
        }
        return res;
    }
    friend Poly operator%(const Poly& lhs, const T modulo) {
        Poly<d, T> res;
        for (int i = 0; i < d; i++) {
            res[i] = lhs[i] % modulo;
            if (res[i] > modulo / 2) res[i] -= modulo;
            else if (res[i] <= -modulo / 2) res[i] += modulo;
        }
        return res;
    }
    friend ostream& operator<<(ostream& os, const Poly& rhs)
    {
        bool prettyPoly = false;
        bool firstPrint = true;
        if (prettyPoly) {
            for (int i = 0; i < d; i++) {
                if (rhs.coeffs[i] != 0) {
                    if (rhs.coeffs[i] > 0 && !firstPrint) os << "+";
                    os << rhs.coeffs[i];
                    if (i > 1) {
                        os << "x^" << i;
                    }
                    else if (i > 0) {
                        os << "x";
                    }
                    firstPrint = false;
                }
            }
            if (firstPrint) os << "0";
        }
        else {
            os << "[" << rhs.coeffs[0];
            for (int i = 1; i < d; i++) {
                os << " " << rhs.coeffs[i];
            }
            os << "]";
        }
        return os;
    }

protected:
    static T _q;
    static stack<T> _qStack;
    static UniformDist uniform_dist;
    std::array<T, d> coeffs;
public:
    Poly() : coeffs() {
        //coeffs[9] = 3;
    }
    Poly(T c) : coeffs() {
        coeffs[0] = c;
    }
    Poly(const initializer_list<T>& list) : coeffs() {
        int i = 0;
        for (T c : list) {
            coeffs[i] = c;
            i++;
        }
    }
    Poly operator-() {
        Poly<d, T> res;
        if(_q == 0)
            for (int i = 0; i < d; i++)
                res[i] = -this->coeffs[i];
        else
            for (int i = 0; i < d; i++) {
                if (coeffs[i] > 0) res[i] = _q - this->coeffs[i];
                else res[i] = -this->coeffs[i];
            }
        return res;
    }
    Poly& operator+=(const Poly& rhs) {
        if (_q == 0)
            for (int i = 0; i < d; i++)
                this->coeffs[i] += rhs.coeffs[i];
        else
            for (int i = 0; i < d; i++)
                this->coeffs[i] = Montgomery::add(this->coeffs[i], rhs.coeffs[i]);
        return *this;
    }
    Poly& operator/=(const T scalar) {
        for (int i = 0; i < d; i++) {
            this->coeffs[i] /= scalar;
        }
        return *this;
    }
    Poly modPos(const T modulo) {
        Poly<d, T> res;
        for (int i = 0; i < d; i++) {
            res[i] = (*this)[i] % modulo;
            if (res[i] < 0) res[i] += modulo;
        }
        return res;
    }
    Poly mod2(const Poly& rhs) {
        Poly<d, T> res = *this;
        int ldeg = res.deg(), rdeg = rhs.deg();
        if (ldeg < rdeg) return res;
        for (int i = ldeg; i >= rdeg; i--) {
            for (int j = 0; j <= rdeg; j++) {
                res[i + j - rdeg] = abs((res[i + j - rdeg] - res[i]*rhs[j]) % 2);
            }
        }
        return res;
    }
    Poly Maugment() {
        Poly<d, T> res;
        if (_q == 0)
            for (int i = 0; i < d; i++)
                res[i] = (*this)[i];
        else
            for (int i = 0; i < d; i++)
                res[i] = Montgomery::augment((*this)[i]);
        return res;
    }
    Poly Mreduce() {
        Poly<d, T> res;
        if (_q == 0)
            for (int i = 0; i < d; i++)
                res[i] = (*this)[i];
        else
            for (int i = 0; i < d; i++)
                res[i] = Montgomery::reduce((*this)[i]);
        return res;
    }
    T& operator[](size_t c)
    {
        return coeffs[c];
    }
    T operator[](size_t c) const {
        return coeffs[c];
    }
    T shift() {
        T temp = coeffs[d];
        for (int i = d - 1; i > 0; i--) {
            coeffs[i] = coeffs[i - 1];
        }
        coeffs[0] = 0;
        return temp;
    }
    T maxAbsCoeff() const {
        T max = 0;
        for (int i = 0; i < d; i++) {
            T coeff = abs((*this)[i]);
            if (coeff > max) max = coeff;
        }
        return max;
    }
    int deg() const {
        int degree = 0;
        for (int i = 0; i < d; i++) {
            if (coeffs[i] != 0) degree = i;
        }
        return degree;
    }
    static Poly One() {
        Poly<d, T> res;
        //if (_q == 0)
        res[0] = 1;
        //else res[0] = Montgomery::augment(1);
        return res;
    }
    static void set_q(T q) {
        if (q != _q) {
            _q = q;
            Montgomery::set_mod(q);
            uniform_dist = UniformDist(q);
            primRootUnity();
        }
    }
    static void push_q(T q) {
        _qStack.push(q);
        //cout << "push length " << _qStack.size() << endl;
        set_q(q);
    }
    static void pop_q() {
        _qStack.pop();
        //cout << "pop length " << _qStack.size() << endl;

        set_q(_qStack.top());
    }

    static Poly Uniform() {
        Poly<d, T> res;
        for (int i = 0; i < d; i++) {
            res[i] = uniform_dist.get();
        }
        return res;
    }
    static Poly Chi() {
        Poly<d, T> res;
        //if(_q == 0)
        for (int i = 0; i < d; i++)
            res[i] = TernaryDist::get();
        /*else
            for(int i = 0; i < d; i++)
                res[i] = Montgomery::augment((T)dist(prng) - 1);*/
        return res;
    }

    static void primRootUnity() {
        //Z pow2 = 1;
        //while (pow2 < 2 * d) pow2 <<= 1;
        //Z a = (_q - 1) / pow2;
        //for (int i = 2; i < _q; i++) {
        //    if (Montgomery::modPow(i, pow2) != 1 && Montgomery::modPow(i, a*pow2/2) != 1) {
        //        primroot = Montgomery::modPow(i, a);
        //        break;
        //    }
        //}
        ////primrootpow2 = pow2;
        primroot = primroots[_q];
        primrootinv = Montgomery::inv(primroot);
    }
    static void reducePoly(array<Z, primrootpow2>& a) {
        if ((phi_m & (phi_m - 1)) == 0) {
            for (int i = primrootpow2 - 1; i >= d; i--) {
                a[i - d] = Montgomery::sub(a[i - d], a[i]);
            }
        }
        else if(phi_m == 625){
            auto phi = phi625;
            for (int i = primrootpow2 - 1; i >= d; i--) {
                for (int j = 0; j < phi.size() - 1; j++) {
                    a[i - d + phi[j]] = Montgomery::sub(a[i - d + phi[j]], a[i]);
                }
            }
        }
        else {
            for (int i = primrootpow2 - 1; i >= d; i--) {
                for (int j = i - d; j <= i - 1; j++) {
                    a[j] = Montgomery::sub(a[j], a[i]);
                }
            }
        }
    }
    static Poly multFFT(const Poly& lhs, const Poly& rhs) {
        array<Z, primrootpow2> a = {}, b = {};
        for (int i = 0; i < d; i++) {
            a[i] = Montgomery::augment(lhs[i]);
            b[i] = Montgomery::augment(rhs[i]);
        }
        DFT(a, false);
        DFT(b, false);
        for (int i = 0; i < primrootpow2; i++) a[i] = Montgomery::mul(a[i], b[i]);
        DFT(a, true);
        Poly<d, Z> res;
        reducePoly(a);
        for (int i = 0; i < d; i++) res[i] = Montgomery::reduce(a[i]);
        return res;
    }
    static Poly multFFTaugmented(array<Z, primrootpow2>& a, array<Z, primrootpow2>& b) {
        array<Z, primrootpow2> c;
        for (int i = 0; i < primrootpow2; i++) c[i] = Montgomery::mul(a[i], b[i]);
        DFT(c, true);
        Poly<d, Z> res;
        for (int i = primrootpow2 - 1; i >= d; i--) {
            c[i - d] = Montgomery::sub(c[i - d], c[i]);
        }
        for (int i = 0; i < d; i++) res[i] = Montgomery::reduce(c[i]);
        return res;
    }
    static void DFT(array<Z, primrootpow2>& a, bool inv) {
        int n = primrootpow2;
        Z one = Montgomery::augment(1);
        for (int i = 1, j = 0; i < n; i++) {
            int bit = n >> 1;
            for (; j & bit; bit >>= 1)
                j ^= bit;
            j ^= bit;

            if (i < j)
                swap(a[i], a[j]);
        }

        for (int len = 2; len <= n; len <<= 1) {
            Z wlen = inv ? primrootinv : primroot;
            wlen = Montgomery::augment(wlen);
            for (int i = len; i < primrootpow2; i <<= 1)
                wlen = Montgomery::mul(wlen, wlen);

            for (int i = 0; i < n; i += len) {
                Z w = one;
                for (int j = 0; j < len / 2; j++) {
                    Z u = a[i + j];
                    Z v = Montgomery::mul(a[i + j + len / 2], w);
                    a[i + j] = Montgomery::add(u, v);
                    a[i + j + len / 2] = Montgomery::sub(u, v);
                    w = Montgomery::mul(w, wlen);
                }
            }
        }

        if (inv) {
            Z n_1 = Montgomery::inv(n);
            n_1 = Montgomery::augment(n_1);
            for (Z& x : a)
                x = Montgomery::mul(x, n_1);
        }
    }
};

template<size_t d, class T>
T Poly<d, T>::_q = 0;
template<size_t d, class T>
stack<T> Poly<d, T>::_qStack;

template<size_t d, class T>
UniformDist Poly<d, T>::uniform_dist;

#endif